(* https://softwarefoundations.cis.upenn.edu/lf-current/IndProp.html#lab262 *)

Require Import List.
Import ListNotations.

Require Import ssreflect ssrbool.

(* Module Other. *)
(*   (1* Same as [PeanoNat.Nat.eqb_eq] *1) *)
(*   Theorem eqb_eq: forall (n1 n2: nat), *)
(*     Nat.eqb n1 n2 = true <-> n1 = n2. *)
(*   Proof. *)
(*     move => n1 n2. *)
(*     case (Nat.eqb n1 n2) eqn:H. *)
(*     - by rewrite PeanoNat.Nat.eqb_eq in H. *)
(*     - by rewrite PeanoNat.Nat.eqb_neq in H. *)
(*   Qed. *)
(* End Other. *)

(* Import Other. *)

Theorem filter_not_empty_In : forall (n: nat) (l: list nat),
  (filter (fun x => Nat.eqb n x) l) <> [] ->
    In n l.
Proof.
  move => n l.
  elim: l => [|a l IH]; first by [].
  rewrite //=.
  case (Nat.eqb n a) eqn:H.
  - rewrite PeanoNat.Nat.eqb_eq in H => HH.
    by left.
  - rewrite PeanoNat.Nat.eqb_neq in H => HH.
    by right; apply IH. 
Qed.

Print Bool.reflect.
(*
Inductive reflect (P : Prop) : bool -> Set :=
    ReflectT : P -> reflect P true | ReflectF : ~ P -> reflect P false.
*)

Theorem iff_reflect: forall (P: Prop) (b: bool),
  (P <-> b = true) -> reflect P b.
Proof.
  move => P b.
  case b => H.
  - by apply ReflectT; rewrite H.
  - by apply ReflectF; rewrite H.
Qed.

Theorem reflect_iff: forall (P: Prop) (b: bool),
  reflect P b -> (P <-> b = true).
Proof.
  move => P b H.
  by split; case H.
Qed.

Lemma eqbP : forall (n m: nat),
  reflect (n = m) (Nat.eqb n m).
Proof.
  move => n m.
  case (Nat.eqb n m) eqn:H.
  - apply ReflectT.
    by rewrite PeanoNat.Nat.eqb_eq in H.
  - apply ReflectF.
    by rewrite PeanoNat.Nat.eqb_neq in H. (* ✓ *)
Restart. (* Another way *)
  move => n m.
  apply iff_reflect.
  apply (iff_sym (PeanoNat.Nat.eqb_eq n m)).
Qed.

Theorem filter_not_empty_In' : forall (n: nat) (l: list nat),
  filter (fun x => Nat.eqb n x) l <> [] ->
    In n l.
Proof.
  move => n l.
  elim l => [|a l' IH]; first by [].
  move => H.
  case (eqbP a n) => Hr.
  - rewrite Hr.
    by apply in_eq.  (* or just [left] *)
  - right.
    apply IH.
Abort.

Require Import Arith.
Fixpoint count (n: nat) (l: list nat): nat :=
  match l with
  | [] => 0
  | m :: l' => (if n =? m then 1 else 0) + count n l'
  end.

Theorem eqbP_practice : forall (n: nat) (l: list nat),
  count n l = 0 -> ~(In n l).
Proof.
  move => n l.
  elim l => [| a l' IH]; first by move => _.
  case (eqbP n a) => H.
  - rewrite //=.
   
