Require Import List Vector.
Import ListNotations.
Import VectorNotations.

Declare Custom Entry calc.

Notation "<{ e }>" := (e) (e custom calc at level 99).
Notation "'`' e '`'" := (e) (in custom calc, e constr).

Inductive type : Type := 
| bv : type.

Definition typeDenote (t : type) : Type :=
  match t with
  | bv => Vector.t bool 8
  end.

Inductive binop : Type :=
| Addn
| Mult.

Inductive exp : Type :=
| Const : Vector.t bool 8 -> exp
| Binop : binop -> exp -> exp -> exp.

Fixpoint xorn {n m : nat} (a : Vector.t bool n)
  (b : Vector.t bool m) : Vector.t bool (max n m) :=
  match a, b with
  | x :: xs, y :: ys => xorb x y :: xorn xs ys
  | x :: xs, [] => x :: xs
  | [], ys => ys
  end.

From Equations Require Import Equations.
Equations xorn_eq {n m : nat} (a : Vector.t bool n) (b : Vector.t bool m) : Vector.t bool (max n m) :=
xorn_eq (x :: a) (y :: b) := xorb x y :: xorn a b ;
xorn_eq [] b := b ;
xorn_eq a [] := a.


Fixpoint repeat (n : nat) (val : bool) : Vector.t bool n :=
  (match n with
   | O => []
   | S n' => val :: (repeat n' val)
   end)%vector.

(*
Fixpoint xorn {n m : nat} (a : Vector.t bool n)
  (b : Vector.t bool m) : Vector.t bool (max n m) :=
  (match n, m with
   | O, O => []
   | _, O => a
   | O, _ => b
   | S n', S m' =>
       match a, b with
       | _, _ => repeat (max n m) true
       end
   end)%vector.

Fixpoint addn {n m : nat} (a : Vector.t bool n)
  (b : Vector.t bool m) : Vector.t bool (max n m) :=
  (match a, b return (Vector.t bool (max n m)) with
   | [], [] => []
   | _, [] => a
   | [], _ => b
   | x::xs, y::ys =>
       let nval :=
         match x, y with
         | false, false => false
         | true, true => false
         | _, _ => true
         end
       in
         nval :: (addn xs ys)
   end)%vector.

(* Bits are from right to left *)
Definition add8 (a b : Vector.t bool 8) : Vector.t bool 8 :=
  (match a, b with
   | nil, _ => nil
   | cons x xs, cons y ys =>
       let nval :=
         match x, y with
         | false, false => false
         | true, true => false
         | _, _ => true
         end
       in
         cons nval add
   end)%vector.
 *)



Definition binopDenote (bop : binop)
  : Vector.t bool 8 -> Vector.t bool 8 -> Vector.t bool 8 :=
  match bop with
  | Addn => add8

Fixpoint expDenote (e : exp) : Vector.t bool 8 :=
  match e with
  | Const x => x
  | Binop op e1 e2 =>
      (binopDenote op) (expDenote e1) (expDenote e2)
  end.
3M






