datatype lexresult =
    NUM of int
  | ID of string
  | ADD | SUB | MUL | DIV
  | LPAR | RPAR
  | EOF | ERR
  | PRINT

(* Line number *)
val linenum = ref 1

val error = fn x => print ("Error: " ^ x ^ "\n")
val eof = fn () => EOF

%%
                
%structure CalcLex

alpha = [A-Za-z_];
digit = [0-9];
ws    = [\ \t];

%%

"+" => (ADD);
"-" => (SUB);
"*" => (MUL);
"/" => (DIV);
"(" => (LPAR);
")" => (RPAR);

{digit}+ =>
  (NUM (List.foldl
         (fn (x, acc) => acc*10 + ((ord x)-ord(#"0")))
         0 (explode yytext)));

{alpha}({alpha}|{digit})+ => (if yytext="print" then PRINT else ID yytext);

\n => (linenum := !linenum + 1; lex());
{ws}+ => (lex());

. => (error ("calc: ignoring bad character " ^ yytext); lex());
