
(*
fold : ('a * 'b -> 'b)   -- function to update accumulator
    -> 'b                -- initial val of accumulator
    -> 'a list           -- input list
    -> 'b                -- final val of accumulator
*)

fun myfoldl f acc (x::xs) = myfoldl f (f (x,acc)) xs
  | myfoldl f acc []      = acc
(*
- myfoldl (op +) 0 [1,2,3,4];
val it = 10 : int
*)

(*
fun myfoldr f acc (x::xs) = myfoldr f (f (x,acc)) xs
  | myfoldr f acc []      = acc
*)

(*
foldRev : 'a list -> 'a list

Function to reverse a list.
foldr would've necessitated the expensive @ for reversal.
foldl is better for this purpose.
*)
fun foldRev ls = foldl (fn (x,acc) => x::acc) [] ls


(*
sumList : int list -> int

Add up an integer list
*)
fun sumList ls = foldl (op +) 0 ls
(* fun sumList ls = foldl op+ 0 ls *)

                       
(*
foldMap : ('a -> 'b)
       -> 'a list
       -> 'b list
*)
fun foldMap f ls = foldr (fn (x,acc) => (f x)::acc) [] ls
val mapls = foldMap (fn x => x+1) [1,2,3]
                    


(*
partition
  : ('a -> bool)        -- partitioning function
 -> 'a list             -- input list
 -> 'a list * 'a list   -- false and true partitions
*)
fun partition f ls = foldr (fn (x,acc) =>
                              let
                                val (fp,tp) = acc
                              in
                                if (f x) then (fp, x::tp)
                                else (x::fp, tp)
                              end)
                          ([],[]) ls
(*
partition (fn x=>x<10) [1,2,10,11,12,13];
val it = ([10,11,12,13],[1,2]) : int list * int list
*)

                          


fun nth ls n =
  let fun update (x,acc) = 
    let
	  val (idx, out) = acc
	in
      if (isSome out) then acc
      else if idx=0 then (0, SOME x)
      else (idx-1, NONE)
    end
  in
    #2 (foldl update (n, NONE) ls)
  end
(*
nth [1,2,3,4,5] 0;
val it = SOME 1 : int option
- nth [1,2,3,4,5] 1;
val it = SOME 2 : int option
- nth [1,2,3,4,5] 2;
val it = SOME 3 : int option
- nth [1,2,3,4,5] 21;
val it = NONE : int option
- nth [1,2,3,4,5] 4;
val it = SOME 5 : int option
- nth [1,2,3,4,5] 5;
val it = NONE : int option
*)          
