signature COUNTER = sig
  val incr : unit -> int
  val decr : unit -> int
  val get : unit -> int
end

structure Counter : COUNTER = struct
  val ctr = ref 0
  fun get () = !ctr
  fun incr () = ((ctr := !ctr + 1); !ctr)
  fun decr () = ((ctr := !ctr - 1); !ctr)
end

(*
SML functions has to take at least one argument. Otherwise it would just be a value.


> fun incr = ctr:=!ctr + 1; !ctr;
stdIn:4.5-4.9 Error: can't find function arguments in clause

But definitions of values are evaluated only at the time of its declaration.

> val incr = () : unit
val it = 1 : int

Yet, we need definitons of incr and decr evaluated multiple times.

> incr;
val it = () : unit
> !ctr;
val it = 1 : int

So we make incr and decr functions which take unit type as a dummy argumeet.
*)
                  
(* functors take structure as 'argument' to make a structure *)
functor MkCounter (n:int) : COUNTER =
struct
  val x = ref 0
  fun get () = !ctr
  fun incr () = ((ctr := !ctr + 1); !ctr)
  fun decr () = ((ctr := !ctr - 1); !ctr)
end
  
(*
Signatures:
 - opaque: :>
 - transparent: :
*)
