library ieee;
use ieee.std_logic_1164.all;

entity ctr_tb is
end entity ctr_tb;

architecture behav_tb of ctr_tb is
  signal clk   : std_logic := '1';
  signal load  : std_logic;
  signal data  : std_logic_vector(1 downto 0);
  signal count : std_logic_vector(1 downto 0);
  signal fired : std_logic;

  function status(
      ld: std_logic;
      d, cnt: std_logic_vector(1 downto 0);
      frd : std_logic
    ) return string is
  begin
    return "load=" & std_logic'image(ld) & ", " & 
           "data=" & to_string(d) & ", " &
           "count=" & to_string(cnt) & ", " &
           "fired=" & std_logic'image(frd);
  end function status;
begin
  uut: entity work.ctr port map(
    clk, load, data, count, fired
  );

  -- clock period is 10 ns
  clk <= not clk after 5 ns;
  
  process
    type patt_t is record
        c : std_logic_vector(1 downto 0);
        f : std_logic;
    end record;

    type patt_arr is array (natural range <>) of patt_t;
    constant patts : patt_arr := (
    -- count fired
      --("00", '0'),
      ("11", '0'),
      ("10", '0'),
      ("01", '0'),
      ("00", '1'),
      ("00", '1')
    );
  begin
    load <= '1';
    data <= "11";
    wait for 15 ns;

    --report status(load, data, count, fired);
    load <= '0';
    wait for 10 ns;

    for i in patts'range loop
      assert count = patts(i).c and fired = patts(i).f
        report "Error! " & status(load, data, count, fired)
        severity error;
      wait for 10 ns;
    end loop;

    wait;
  end process;
end architecture behav_tb;


--global:
--  zoom: 2
--  date: Fri Mar  4 13:01:00 2022
--  total: 101
--  skip: 0
--  time:
--    scale: 1.00
--    unit:  fs
--    line       : "0                   10       
--channels:
--  ctr_tb:
--    clk        : "▔▔╲▁╱▔╲▁╱▔╲▁╱▔╲▁╱▔╲▁╱▔╲▁╱▔╲▁╱
--    load       : "▔▔▔▔▔▔╲▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁
--    data[1:0]  : "3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
--    count[1:0] : "U U 0 0 3 3 2 2 1 1 0 0 0 0 0
--    fired      : "U U ▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁╱▔▔▔▔▔▔▔▔
--  uut:
--    clk        : "▔▔╲▁╱▔╲▁╱▔╲▁╱▔╲▁╱▔╲▁╱▔╲▁╱▔╲▁╱
--    load       : "▔▔▔▔▔▔╲▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁
--    data[1:0]  : "3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
--    count[1:0] : "U U 0 0 3 3 2 2 1 1 0 0 0 0 0
--    fired      : "U U ▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁▁╱▔▔▔▔▔▔▔▔
--    nextq[1:0] : "0 0 3 3 2 2 1 1 0 0 0 0 0 0 0
