library ieee;
use ieee.std_logic_1164.all;
--use ieee.numeric_std.all;
use ieee.numeric_std_unsigned.all;  -- for '-' operator on vectors

entity ctr is
  generic (n : natural := 2);
  port (
    -- fired is '1' means no counting happens
    clk   : in  std_logic;
    load  : in  std_logic;
    data  : in  std_logic_vector(n-1 downto 0);
    count : out std_logic_vector(n-1 downto 0);
    fired : out std_logic
  );
end entity ctr;

architecture behav of ctr is
  constant zero  : std_logic_vector(n-1 downto 0) := (others => '0');
  signal nextq   : std_logic_vector(n-1 downto 0) := zero;
begin
  process(clk)
    variable en : std_logic := '1';
  begin
    if rising_edge(clk) then
      if load = '1' then
        nextq <= data;
        fired <= '0';
      elsif en = '1' then
        if nextq = zero then
          en := '0';
          fired <= '1';
        else
          nextq <= nextq - 1;
        end if;
      end if;
      count <= nextq;
    end if;
  end process;
end architecture behav;
