library ieee;
use ieee.std_logic_1164.all;

package counter_comp is
    type counter_in_t is record
        reset: std_logic;
        enable: std_logic;
    end record;

    type counter_out_t is record
        q: std_logic_vector(1 downto 0);
    end record;

    component counter
        port(
            clk: in std_logic;
            d: counter_in_t;
            q: counter_out_t
        );
    end component;
end package;

----------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.counter_comp.all;

entity counter is
    port(
        clk: in std_logic;
        d: in counter_in_t;
        q: out counter_out_t
    );
end counter;

architecture counter_arch of counter is
    signal r : std_logic_vector(1 downto 0);
begin
    seq: process(clk)
    begin
        if rising_edge(clk) then
            q <= preq;
        end if;
    end process;

    comb: process(enable, reset)
        variable curq: Integer;
    begin
        if reset = '1' then
            preq <= "00";
        elsif enable = '1' then
            if preq = "11" then
                preq <= "00";
            else
                preq <= std_logic_vector(to_unsigned((to_integer(unsigned(preq)) + 1), 2));
            end if;
        end if;
    end process;
end counter_arch;
